/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2019/20      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : Assignment 2
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description:
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <cassert>
#include <sstream>
#include <chrono>
#include <thread>
#include <atomic>
#include <mutex>

//include,files
#include "barrier.h"


typedef std::chrono::milliseconds TimeT;

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Read transition matrix data from file
////////////////////////////////////////////////////////////////////////////////////////////////////
void
readMat( const std::string& fname, const int nmat, std::vector<double>& mat) {

  // open file
  std::ifstream file( fname, std::ios::binary);
  if( ! file.good()) {
    std::cerr << "readMat() : Failed to open '" << fname << "'." << std::endl;
    exit(0);
  }

  // determine file size
  file.seekg( 0, std::ios::end);
  auto fsize = file.tellg();
  file.seekg( 0);
  assert( fsize / sizeof(double) == nmat * nmat);

  // read data
  mat.resize( fsize / sizeof(double));
  file.read( (char*) mat.data(), fsize);

  file.close();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Verify stationary vector
////////////////////////////////////////////////////////////////////////////////////////////////////
double
verifyStationary( const int nmat, std::vector<double>& mat, std::vector<double>& stationary) {

  // right matrix-vector product
  std::vector<double> temp( nmat, 0.);
  for( int i = 0; i < nmat; ++i) {
    for( int j = 0; j < nmat; ++j) {
      temp[i] += mat[j*nmat+i] * stationary[j];
    }
  }

  // compute L_1 residual
  double err = 0.;
  for( int i = 0; i < (int) temp.size(); ++i) {
    err += std::abs( temp[i] - stationary[i]);
  }

  return err;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//! Implementation of power iteration
//! @num_iterations  number of iterations
//! @nmat  size of matrix
//! @mat   matrix (row-major order)
//! @stationary  stationary vector to be computed
////////////////////////////////////////////////////////////////////////////////////////////////////
double
powerIterationSerial( const int num_iterations, const int nmat,
                      std::vector<double>& mat, std::vector<double>& stationary)
{

  auto start = std::chrono::steady_clock::now();

  // TODO: implement power iteration

  double time = std::chrono::duration_cast<TimeT>(std::chrono::steady_clock::now()-start).count();

  return time;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Parallel implementation of power iteration
//! @num_iterations  number of iterations
//! @nmat  size of matrix
//! @mat   matrix (row-major order)
//! @stationary  stationary vector to be computed
////////////////////////////////////////////////////////////////////////////////////////////////////
double
powerIterationParallel( const int num_iterations, const int nmat,
                        std::vector<double>& mat, std::vector<double>& stationary)
{

  auto start = std::chrono::steady_clock::now();

  // TODO: implement power iteration in parallel

  double time = std::chrono::duration_cast<TimeT>(std::chrono::steady_clock::now()-start).count();

  return time;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Parallel implementation of power iteration
//! @num_iterations  number of iterations
//! @nmat  size of matrix
//! @mat   matrix (row-major order)
//! @stationary  stationary vector to be computed
////////////////////////////////////////////////////////////////////////////////////////////////////
double
powerIterationParallelBarrier( const int num_iterations, const int nmat,
                               std::vector<double>& mat, std::vector<double>& stationary)
{

  auto start = std::chrono::steady_clock::now();

  // TODO: implement power iteration in parallel without restarting threads every iteration

  double time = std::chrono::duration_cast<TimeT>(std::chrono::steady_clock::now()-start).count();

  return time;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//!
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/) {

  // grid resolution to use
  const int ntheta = 64;
  const int nphi = 128;
  const int nmat = ntheta * nphi;

  // load transition matrix
  std::vector<double> mat;
  std::stringstream fname;
  fname << "./data/mat_" << ntheta << "x" << nphi << ".dat";
  readMat( fname.str(), nmat, mat);

  std::vector<double> stationary;
  double timeserial = powerIterationSerial( 50, nmat, mat, stationary);
  std::cout << "time serial = " << timeserial << " ms." << std::endl;

  double timeparallel = powerIterationParallel( 50, nmat, mat, stationary);
  std::cout << "time parallel = " << timeparallel << " ms." << std::endl;

  std::vector<double> stationary_barrier;
  double timebarrier = powerIterationParallelBarrier( 50, nmat, mat, stationary_barrier);
  std::cout << "time barrier = " << timebarrier << " ms." << std::endl;

  double err = verifyStationary( nmat, mat, stationary_barrier);
  std::cout << "err = " << err << std::endl;
  // TODO: suitable threshold to determine if err is admissible


}
